#!/usr/bin/env python
# -*- coding: ISO-8859-1 -*-
# pylint: disable=R0913,W0702,R0914,R0912,R0201
"""
File: pycurl_manager.py
Author: Valentin Kuznetsov <vkuznet@gmail.com>
Description: a basic wrapper around pycurl library.
The RequestHandler class provides basic APIs to get data
from a single resource or submit mutliple requests to
underlying data-services.

Examples:
# CERN SSO: http://linux.web.cern.ch/linux/docs/cernssocookie.shtml
# use RequestHandler with CERN SSO enabled site
mgr = RequestHandler()
url = "https://cms-gwmsmon.cern.ch/prodview/json/site_summary"
params = {}
tfile = tempfile.NamedTemporaryFile()
cern_sso_cookie(url, tfile.name, cert, ckey)
cookie = {url: tfile.name}
header, data = mgr.request(url3, params, cookie=cookie)
if header.status != 200:
    print "ERROR"

# fetch multiple urls at onces from various urls
tfile = tempfile.NamedTemporaryFile()
ckey = os.path.join(os.environ['HOME'], '.globus/userkey.pem')
cert = os.path.join(os.environ['HOME'], '.globus/usercert.pem')
url1 = "https://cmsweb.cern.ch/dbs/prod/global/DBSReader/help"
url2 = "https://cmsweb.cern.ch/dbs/prod/global/DBSReader/datatiers"
url3 = "https://cms-gwmsmon.cern.ch/prodview/json/site_summary"
cern_sso_cookie(url3, tfile.name, cert, ckey)
cookie = {url3: tfile.name}
urls = [url1, url2, url3]
data = getdata(urls, ckey, cert, cookie=cookie)
for row in data:
    print(row)
"""
from __future__ import print_function
from future import standard_library
standard_library.install_aliases()

from builtins import str, range, object
from past.builtins import basestring
from future.utils import viewitems

# system modules
import json
import logging
import os
import re
import subprocess
import sys
import pycurl
from io import BytesIO
import http.client
from urllib.parse import urlencode

class ResponseHeader(object):
    """ResponseHeader parses HTTP response header"""

    def __init__(self, response):
        super(ResponseHeader, self).__init__()
        self.header = {}
        self.reason = ''
        self.fromcache = False
        self.parse(response)

    def parse(self, response):
        """Parse response header and assign class member data"""
        startRegex = r"^HTTP/\d.\d \d{3}"
        continueRegex = r"^HTTP/\d.\d 100"  # Continue: client should continue its request
        replaceRegex = r"^HTTP/\d.\d"

        for row in response.split('\r'):
            row = row.replace('\n', '')
            if not row:
                continue
            if re.search(startRegex, row):
                if re.search(continueRegex, row):
                    continue
                res = re.sub(replaceRegex, "", row).strip()
                status, reason = res.split(' ', 1)
                self.status = int(status)
                self.reason = reason
                continue
            try:
                key, val = row.split(':', 1)
                self.header[key.strip()] = val.strip()
            except:
                pass

    def getReason(self):
        """
        Return the HTTP request reason
        """
        return self.reason

    def getHeader(self):
        """
        Return the header dictionary object
        """
        return self.header

    def getHeaderKey(self, keyName):
        """
        Provided a key name, return it from the HTTP header.
        Note that - by design - header keys are meant to be
        case insensitive
        :param keyName: a header key name to be looked up
        :return: the value for that header key, or None if not found
        """
        for keyHea, valHea in self.header.items():
            if keyHea.lower() == keyName.lower():
                return valHea


class RequestHandler(object):
    """
    RequestHandler provides APIs to fetch single/multiple
    URL requests based on pycurl library
    """

    def __init__(self, config=None, logger=None):
        super(RequestHandler, self).__init__()
        if not config:
            config = {}
        defaultOpts = pycurl_options()
        self.nosignal = config.get('nosignal', defaultOpts['NOSIGNAL'])
        self.timeout = config.get('timeout', defaultOpts['TIMEOUT'])
        self.connecttimeout = config.get('connecttimeout', defaultOpts['CONNECTTIMEOUT'])
        self.followlocation = config.get('followlocation', defaultOpts['FOLLOWLOCATION'])
        self.maxredirs = config.get('maxredirs', defaultOpts['MAXREDIRS'])
        self.logger = logger if logger else logging.getLogger()

    def encode_params(self, params, verb, doseq, encode):
        """ Encode request parameters for usage with the 4 verbs.
            Assume params is already encoded if it is a string and
            uses a different encoding depending on the HTTP verb
            (either json.dumps or urllib.urlencode)
        """
        if not encode:
            return params
        # data is already encoded, just return it
        if isinstance(params, basestring):
            return params

        # data is not encoded, we need to do that
        if verb in ['GET', 'HEAD']:
            if params:
                encoded_data = urlencode(params, doseq=doseq)
            else:
                return ''
        else:
            if params:
                encoded_data = json.dumps(params)
            else:
                return {}

        return encoded_data

    def set_opts(self, curl, url, params, headers,
                 ckey=None, cert=None, capath=None, verbose=None,
                 verb='GET', doseq=True, encode=False, cainfo=None, cookie=None):
        """Set options for given curl object, params should be a dictionary"""
        if not (isinstance(params, (dict, basestring)) or params is None):
            raise TypeError("pycurl parameters should be passed as dictionary or an (encoded) string")
        curl.setopt(pycurl.NOSIGNAL, self.nosignal)
        curl.setopt(pycurl.TIMEOUT, self.timeout)
        curl.setopt(pycurl.CONNECTTIMEOUT, self.connecttimeout)
        curl.setopt(pycurl.FOLLOWLOCATION, self.followlocation)
        curl.setopt(pycurl.MAXREDIRS, self.maxredirs)

        # also accepts encoding/compression algorithms
        if headers and headers.get("Accept-Encoding"):
            if isinstance(headers["Accept-Encoding"], basestring):
                curl.setopt(pycurl.ENCODING, headers["Accept-Encoding"])
            else:
                logging.warning("Wrong data type for header 'Accept-Encoding': %s",
                                type(headers["Accept-Encoding"]))

        if cookie and url in cookie:
            curl.setopt(pycurl.COOKIEFILE, cookie[url])
            curl.setopt(pycurl.COOKIEJAR, cookie[url])

        encoded_data = self.encode_params(params, verb, doseq, encode)


        if verb == 'GET':
            if encoded_data:
                url = url + '?' + encoded_data
        elif verb == 'HEAD':
            if encoded_data:
                url = url + '?' + encoded_data
            curl.setopt(pycurl.CUSTOMREQUEST, verb)
            curl.setopt(pycurl.HEADER, 1)
            curl.setopt(pycurl.NOBODY, True)
        elif verb == 'POST':
            curl.setopt(pycurl.POST, 1)
            if encoded_data:
                curl.setopt(pycurl.POSTFIELDS, encoded_data)
        elif verb == 'DELETE' or verb == 'PUT':
            curl.setopt(pycurl.CUSTOMREQUEST, verb)
            curl.setopt(pycurl.HTTPHEADER, ['Transfer-Encoding: chunked'])
            if encoded_data:
                curl.setopt(pycurl.POSTFIELDS, encoded_data)
        else:
            raise Exception('Unsupported HTTP method "%s"' % verb)

        if verb in ('POST', 'PUT'):
            # only these methods (and PATCH) require this header
            headers["Content-Length"] = str(len(encoded_data))

        # we must pass url as a string data-type, otherwise pycurl will fail with error
        # TypeError: invalid arguments to setopt
        # see https://curl.haxx.se/mail/curlpython-2007-07/0001.html
        curl.setopt(pycurl.URL, str(url))
        if headers:
            curl.setopt(pycurl.HTTPHEADER, \
                    ["%s: %s" % (k, v) for k, v in viewitems(headers)])
        bbuf = BytesIO()
        hbuf = BytesIO()
        curl.setopt(pycurl.WRITEFUNCTION, bbuf.write)
        curl.setopt(pycurl.HEADERFUNCTION, hbuf.write)
        if capath:
            curl.setopt(pycurl.CAPATH, capath)
            curl.setopt(pycurl.SSL_VERIFYPEER, True)
            if cainfo:
                curl.setopt(pycurl.CAINFO, cainfo)
        else:
            curl.setopt(pycurl.SSL_VERIFYPEER, False)
        if ckey:
            curl.setopt(pycurl.SSLKEY, ckey)
        if cert:
            curl.setopt(pycurl.SSLCERT, cert)
        if verbose:
            curl.setopt(pycurl.VERBOSE, True)
            curl.setopt(pycurl.DEBUGFUNCTION, self.debug)
        return bbuf, hbuf

    def debug(self, debug_type, debug_msg):
        """Debug callback implementation"""
        print("debug(%d): %s" % (debug_type, debug_msg))

    def parse_body(self, data, decode=False):
        """
        Parse body part of URL request (by default use json).
        This method can be overwritten.
        """
        if decode:
            try:
                res = json.loads(data)
                return res
            except ValueError as exc:
                msg = 'Unable to load JSON data, %s, data type=%s, pass as is' \
                      % (str(exc), type(data))
                logging.debug(msg)
                return data
        else:
            return data

    def parse_header(self, header):
        """
        Parse response header.
        This method can be overwritten.
        """
        return ResponseHeader(header)

    def request(self, url, params, headers=None, verb='GET',
                verbose=0, ckey=None, cert=None, capath=None,
                doseq=True, encode=False, decode=False, cainfo=None, cookie=None):
        """Fetch data for given set of parameters"""
        curl = pycurl.Curl()
        bbuf, hbuf = self.set_opts(curl, url, params, headers, ckey, cert, capath,
                                   verbose, verb, doseq, encode, cainfo, cookie)
        curl.perform()
        if verbose:
            print(verb, url, params, headers)
        header = self.parse_header(hbuf.getvalue())
        if header.status < 300:
            if verb == 'HEAD':
                data = ''
            else:
                data = self.parse_body(bbuf.getvalue(), decode)
        else:
            data = bbuf.getvalue()
            msg = 'url=%s, code=%s, reason=%s, headers=%s' \
                  % (url, header.status, header.reason, header.header)
            exc = http.client.HTTPException(msg)
            setattr(exc, 'req_data', params)
            setattr(exc, 'req_headers', headers)
            setattr(exc, 'url', url)
            setattr(exc, 'result', data)
            setattr(exc, 'status', header.status)
            setattr(exc, 'reason', header.reason)
            setattr(exc, 'headers', header.header)
            bbuf.flush()
            hbuf.flush()
            raise exc

        bbuf.flush()
        hbuf.flush()
        return header, data

    def getdata(self, url, params, headers=None, verb='GET',
                verbose=0, ckey=None, cert=None, doseq=True,
                encode=False, decode=False, cookie=None):
        """Fetch data for given set of parameters"""
        _, data = self.request(url=url, params=params, headers=headers, verb=verb,
                               verbose=verbose, ckey=ckey, cert=cert, doseq=doseq,
                               encode=encode, decode=decode, cookie=cookie)
        return data

    def getheader(self, url, params, headers=None, verb='GET',
                  verbose=0, ckey=None, cert=None, doseq=True):
        """Fetch HTTP header"""
        header, _ = self.request(url, params, headers, verb,
                                 verbose, ckey, cert, doseq=doseq)
        return header

    def multirequest(self, url, parray, headers=None,
                     ckey=None, cert=None, verbose=None, cookie=None):
        """Fetch data for given set of parameters"""
        multi = pycurl.CurlMulti()
        for params in parray:
            curl = pycurl.Curl()
            bbuf, hbuf = \
                self.set_opts(curl, url, params, headers, ckey=ckey, cert=cert,
                              verbose=verbose, cookie=cookie)
            multi.add_handle(curl)
            while True:
                ret, num_handles = multi.perform()
                if ret != pycurl.E_CALL_MULTI_PERFORM:
                    break
            while num_handles:
                ret = multi.select(1.0)
                if ret == -1:
                    continue
                while True:
                    ret, num_handles = multi.perform()
                    if ret != pycurl.E_CALL_MULTI_PERFORM:
                        break
            dummyNumq, response, dummyErr = multi.info_read()
            for dummyCobj in response:
                data = json.loads(bbuf.getvalue())
                if isinstance(data, dict):
                    data.update(params)
                    yield data
                if isinstance(data, list):
                    for item in data:
                        if isinstance(item, dict):
                            item.update(params)
                            yield item
                        else:
                            err = 'Unsupported data format: data=%s, type=%s' \
                                  % (item, type(item))
                            raise Exception(err)
                bbuf.flush()
                hbuf.flush()


HTTP_PAT = re.compile( \
    "(https|http)://[-A-Za-z0-9_+&@#/%?=~_|!:,.;]*[-A-Za-z0-9+&@#/%=~_|]")


def validate_url(url):
    "Validate URL"
    if HTTP_PAT.match(url):
        return True
    return False


def pycurl_options():
    "Default set of options for pycurl"
    opts = {
        'FOLLOWLOCATION': 1,
        'CONNECTTIMEOUT': 270,
        'MAXREDIRS': 5,
        'NOSIGNAL': 1,
        'TIMEOUT': 270,
        'SSL_VERIFYPEER': False,
        'VERBOSE': 0
    }
    return opts


def cern_sso_cookie(url, fname, cert, ckey):
    "Obtain cern SSO cookie and store it in given file name"
    cmd = 'cern-get-sso-cookie -cert %s -key %s -r -u %s -o %s' \
          % (cert, ckey, url, fname)
    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True, env=os.environ)
    proc.wait()


def getdata(urls, ckey, cert, headers=None, options=None, num_conn=50, cookie=None):
    """
    Get data for given list of urls, using provided number of connections
    and user credentials
    """

    if not options:
        options = pycurl_options()

    # Make a queue with urls
    queue = [u for u in urls if validate_url(u)]

    # Check args
    num_urls = len(queue)
    num_conn = min(num_conn, num_urls)

    # Pre-allocate a list of curl objects
    mcurl = pycurl.CurlMulti()
    mcurl.handles = []
    for _ in range(num_conn):
        curl = pycurl.Curl()
        curl.fp = None
        for key, val in viewitems(options):
            curl.setopt(getattr(pycurl, key), val)
        curl.setopt(pycurl.SSLKEY, ckey)
        curl.setopt(pycurl.SSLCERT, cert)
        mcurl.handles.append(curl)
        if headers:
            curl.setopt(pycurl.HTTPHEADER, \
                        ["%s: %s" % (k, v) for k, v in viewitems(headers)])

    # Main loop
    freelist = mcurl.handles[:]
    num_processed = 0
    while num_processed < num_urls:
        # If there is an url to process and a free curl object,
        # add to multi-stack
        while queue and freelist:
            url = queue.pop(0)
            curl = freelist.pop()
            curl.setopt(pycurl.URL, url.encode('ascii', 'ignore'))
            if cookie and url in cookie:
                curl.setopt(pycurl.COOKIEFILE, cookie[url])
                curl.setopt(pycurl.COOKIEJAR, cookie[url])
            bbuf = BytesIO()
            hbuf = BytesIO()
            curl.setopt(pycurl.WRITEFUNCTION, bbuf.write)
            curl.setopt(pycurl.HEADERFUNCTION, hbuf.write)
            mcurl.add_handle(curl)
            # store some info
            curl.hbuf = hbuf
            curl.bbuf = bbuf
            curl.url = url
        # Run the internal curl state machine for the multi stack
        while True:
            ret, _ = mcurl.perform()
            if ret != pycurl.E_CALL_MULTI_PERFORM:
                break
        # Check for curl objects which have terminated, and add them to the
        # freelist
        while True:
            num_q, ok_list, err_list = mcurl.info_read()
            for curl in ok_list:
                if sys.version.startswith('3.'):
                    hdrs = curl.hbuf.getvalue().decode('utf-8')
                    data = curl.bbuf.getvalue().decode('utf-8')
                else:
                    hdrs = curl.hbuf.getvalue()
                    data = curl.bbuf.getvalue()
                url = curl.url
                curl.bbuf.flush()
                curl.bbuf.close()
                curl.hbuf.close()
                curl.hbuf = None
                curl.bbuf = None
                mcurl.remove_handle(curl)
                freelist.append(curl)
                yield {'url': url, 'data': data, 'headers': hdrs}
            for curl, errno, errmsg in err_list:
                hdrs = curl.hbuf.getvalue()
                data = curl.bbuf.getvalue()
                url = curl.url
                curl.bbuf.flush()
                curl.bbuf.close()
                curl.hbuf.close()
                curl.hbuf = None
                curl.bbuf = None
                mcurl.remove_handle(curl)
                freelist.append(curl)
                yield {'url': url, 'data': None, 'headers': hdrs, \
                       'error': errmsg, 'code': errno}
            num_processed = num_processed + len(ok_list) + len(err_list)
            if num_q == 0:
                break
        # Currently no more I/O is pending, could do something in the meantime
        # (display a progress bar, etc.).
        # We just call select() to sleep until some more data is available.
        mcurl.select(1.0)

    cleanup(mcurl)


def cleanup(mcurl):
    "Clean-up MultiCurl handles"
    for curl in mcurl.handles:
        if curl.hbuf is not None:
            curl.hbuf.close()
            curl.hbuf = None
        if curl.bbuf is not None:
            curl.bbuf.close()
            curl.bbuf = None
        curl.close()
    mcurl.close()



#!/usr/bin/env python
"""
_DASDocument_

A class representing the structure of data in DAS. The cannonical description of
this structure is found on:
        https://twiki.cern.ch/twiki/bin/view/CMS/DMWMDataAggregationService
"""


class DASDocument(dict):
    def __init__(self, service):
        """
        Make an empty dictionary representing a DAS document
        """
        dict = {"request_timestamp": 0,
                "request_url": "",
                "request_method": "GET",
                "request_params": {},
                "response_version": 0,
                "response_expires": 0,
                "response_checksum": "",
                "request_api": "",
                "call_time": 0}
        self.keys = dict.keys()
        self.setdefault(service, dict)

    def compare(self, dict):
        """
        Is the provided dict a valid DAS document
        """
        pass

    def validate(self):
        """
        Is this instance of a DAS document valid?
        """

        return True



#!/usr/bin/env python
"""
Script meant to be used to check the status of a dataset (or lfn) against DBS
and PhEDEx. It also prints any discrepancy between those 2 data management tools.
"""
from __future__ import print_function, division
from future import standard_library
standard_library.install_aliases()

import http.client
import json
import os
import sys
import urllib.request, urllib.parse
from urllib.error import HTTPError, URLError
from argparse import ArgumentParser

main_url = "https://cmsweb.cern.ch"
phedex_url = main_url + "/phedex/datasvc/json/prod/"
dbs_url = main_url + "/dbs/prod/global/DBSReader/"


class HTTPSClientAuthHandler(urllib.request.HTTPSHandler):
    def __init__(self):
        urllib.request.HTTPSHandler.__init__(self)
        self.key = os.getenv("X509_USER_PROXY")
        self.cert = os.getenv("X509_USER_PROXY")

    def https_open(self, req):
        return self.do_open(self.getConnection, req)

    def getConnection(self, host, timeout=300):
        return http.client.HTTPSConnection(host, key_file=self.key, cert_file=self.cert)


def get_content(url, params=None):
    opener = urllib.request.build_opener(HTTPSClientAuthHandler())
    try:
        if params:
            response = opener.open(url, params)
            output = response.read()
        else:
            response = opener.open(url)
            output = response.read()
    except HTTPError as e:
        print('The server couldn\'t fulfill the request. Erro code: ', e.code)
        sys.exit(1)
    except URLError as e:
        print('Failed to reach server. Reason:', e.reason)
        sys.exit(1)
    return output


def phedex_info(dataset):
    """
    Query blockreplicas PhEDEx API to retrieve detailed information
    for a specific dataset
    """
    api_url = phedex_url + "blockreplicas" + "?" + urllib.parse.urlencode([('dataset', dataset)])
    phedex_summary = json.loads(get_content(api_url))
    return phedex_summary


def dbs_info(dataset):
    """
    Queries 2 DBS APIs to get both summary and detailed information
    """
    dbs_out = {}
    api_url = dbs_url + "blocksummaries" + "?" + urllib.parse.urlencode({'dataset': dataset})
    dbs_out['blocksummaries'] = json.loads(get_content(api_url))
    api_url = dbs_url + "files" + "?" + urllib.parse.urlencode({'dataset': dataset}) + "&detail=1"
    dbs_out['files'] = json.loads(get_content(api_url))
    return dbs_out


def main(argv=None):
    """
    Receive either a dataset name or a logical file name
    and proxy location. Then it queries the following data
    services:
     - phedex : gets number of files
     - dbs    : gets the number of valid, invalid and total files

    It returns the number of files for this dataset/lfn available
    in PhEDEx and DBS
    """
    usage = "usage: %prog -d dataset_name"
    parser = ArgumentParser(usage=usage)
    parser.add_argument('-d', '--dataset', help='Dataset name', dest='dataset')
    parser.add_argument('-l', '--lfn', help='Logical file name', dest='lfn')
    options = parser.parse_args()
    if not (options.dataset or options.lfn):
        parser.error("Please supply either dataset name or file name \
                      and certificate location")
        sys.exit(1)
    if options.dataset:
        dataset = options.dataset
    if options.lfn:
        lfn = options.lfn
        lfnAux = lfn.split('/')
        dataset = '/' + lfnAux[4] + '/' + lfnAux[3] + '-' + lfnAux[6] + '/' + lfnAux[5]

    print("Dataset: %s" % dataset)

    phedex_out = phedex_info(dataset)
    dbs_out = dbs_info(dataset)
    phedex_files = 0
    phedex_blocks = {}
    for item in phedex_out["phedex"]["block"]:
        phedex_files += item['files']
        phedex_blocks.setdefault(item['name'], item['files'])

    dbs_files = dbs_out['blocksummaries'][0]['num_file']
    dbs_blocks = {}
    dbs_file_valid = 0
    dbs_file_invalid = 0
    for item in dbs_out['files']:
        dbs_blocks.setdefault(item['block_name'], 0)
        dbs_blocks[item['block_name']] += 1
        if item['is_file_valid']:
            dbs_file_valid += 1
        else:
            dbs_file_invalid += 1

    print("Phedex file count : ", phedex_files)
    print("DBS file count    : ", dbs_files)
    print(" - valid files    : ", dbs_file_valid)
    print(" - invalid files  : ", dbs_file_invalid)
    print(" - valid+invalid  : ", (dbs_file_valid + dbs_file_invalid))
    print("Blocks in PhEDEx but not in DBS: ", set(phedex_blocks.keys()) - set(dbs_blocks.keys()))
    print("Blocks in DBS but not in PhEDEx: ", set(dbs_blocks.keys()) - set(phedex_blocks.keys()))

    for blockname in phedex_blocks:
        if phedex_blocks[blockname] != dbs_blocks.get(blockname):
            print("Block with file mismatch: %s" % blockname)
            print("\tPhEDEx: %s\t\tDBS: %s" % (phedex_blocks.get(blockname), dbs_blocks.get(blockname)))


if __name__ == "__main__":
    sys.exit(main())



"""
_DBFormatter_

A class to define the standardised formatting of database results.
Holds a bunch of helper methods to format input and output of sql
interactions.
"""

import datetime
import time
import types

from WMCore.DataStructs.WMObject import WMObject


class DBFormatter(WMObject):
    def __init__(self, logger, dbinterface):
        """
        The class holds a connection to the database in self.dbi. This is a
        DBCore.DBInterface object.
        """
        self.logger = logger
        self.dbi = dbinterface

    def truefalse(self, value):
        if value in ('False', 'FALSE', 'n', 'N', 'NO', 'No'):
            value = 0
        return bool(value)

    def convertdatetime(self, t):
        return int(time.mktime(t.timetuple()))

    def timestamp(self):
        """
        generate a timestamp
        """
        t = datetime.datetime.now()
        return self.convertdatetime(t)

    def format(self, result):
        """
        Some standard formatting, put all records into a list.
        Returns a list of lists
        """
        out = []
        for r in result:
            for i in r.fetchall():
                row = []
                for j in i:
                    row.append(j)
                out.append(row)
            r.close()
        return out

    def formatOne(self, result):
        """
        Return one record
        """
        out = []
        for r in result:
            for i in r.fetchone():
                out.append(i)
        return out

    def formatDict(self, result):
        """
        Returns an array of dictionaries representing the results
        """
        dictOut = []
        for r in result:
            descriptions = r.keys
            for i in r.fetchall():
                # WARNING: this can generate errors for some stupid reason
                # in both oracle and mysql.
                entry = {}
                for index in xrange(0, len(descriptions)):
                    # WARNING: Oracle returns table names in CAP!
                    if isinstance(i[index], unicode):
                        entry[str(descriptions[index].lower())] = str(i[index])
                    else:
                        entry[str(descriptions[index].lower())] = i[index]

                dictOut.append(entry)

            r.close()

        return dictOut

    def formatList(self, result):
        """
        Returns a flat array with the results.
        Ideally used for single column queries
        """
        listOut = []
        for r in result:
            descriptions = r.keys
            for i in r.fetchall():
                for index in xrange(0, len(descriptions)):
                    if isinstance(i[index], unicode):
                        listOut.append(str(i[index]))
                    else:
                        listOut.append(i[index])
            r.close()
        return listOut

    def formatOneDict(self, result):
        """
        Return a dictionary representing the first record
        """
        if (len(result) == 0):
            return {}

        r = result[0]
        description = [str(x).lower() for x in r.keys]
        if len(r.data) < 1:
            return {}

        return dict(list(zip(description, r.fetchone())))

    def formatCursor(self, cursor, size=10):
        """
        Fetch the driver cursor directly.
        Tested only with cx_Oracle.
        Cursor must be already executed.
        Use fetchmany(size = default arraysize = 50)

        """
        if isinstance(cursor.keys, types.MethodType):
            keys = [x.lower() for x in cursor.keys()]
        else:
            keys = [x.lower() for x in cursor.keys]
        result = []
        while True:
            if not cursor.closed:
                rows = cursor.fetchmany(size=size)
                if not rows:
                    cursor.close()
                    break
                for r in rows:
                    result.append(dict(list(zip(keys, r))))
            else:
                break
        if not cursor.closed:
            cursor.close()
        return result

    def getBinds(self, **kwargs):
        binds = {}
        for i in kwargs.keys():
            binds = self.dbi.buildbinds(self.dbi.makelist(kwargs[i]), i, binds)
        return binds

    def execute(self, conn=None, transaction=False, returnCursor=False):
        """
        A simple select with no binds/arguments is the default
        """
        result = self.dbi.processData(self.sql, self.getBinds(),
                                      conn=conn, transaction=transaction,
                                      returnCursor=returnCursor)
        return self.format(result)

    def executeOne(self, conn=None, transaction=False, returnCursor=False):
        """
        A simple select with no binds/arguments is the default
        """
        result = self.dbi.processData(self.sql, self.getBinds(),
                                      conn=conn, transaction=transaction,
                                      returnCursor=returnCursor)
        return self.formatOne(result)



#!/usr/bin/python
"""
_WMException_

General Exception class for WM modules

"""

from builtins import bytes
from future.utils import viewitems

import inspect
import logging
import sys
import traceback

WMEXCEPTION_START_STR = "<@========== WMException Start ==========@>"
WMEXCEPTION_END_STR = "<@---------- WMException End ----------@>"


class WMException(Exception):
    """
    _WMException_

    Exception class which works out details of where
    it was raised.

    """

    def __init__(self, message, errorNo=None, **data):
        self.name = str(self.__class__.__name__)
        if type(message) == bytes:
            # Fix for the unicode encoding issue, see #8056 and #8403
            # interprets this string using utf-8 codec and ignoring any errors
            message = message.decode('utf-8', 'ignore')

        Exception.__init__(self, self.name, message)

        #  //
        # // Init data dictionary with defaults
        # //
        self.data = {}
        self.data.setdefault("ClassName", None)
        self.data.setdefault("ModuleName", None)
        self.data.setdefault("MethodName", None)
        self.data.setdefault("ClassInstance", None)
        self.data.setdefault("FileName", None)
        self.data.setdefault("LineNumber", None)
        if errorNo is None:
            self.data.setdefault("ErrorNr", 0)
        else:
            self.data.setdefault("ErrorNr", errorNo)

        self._message = message
        self.data.update(data)

        #  //
        # // Automatically determine the module name
        # //  if not set
        if self.data['ModuleName'] is None:
            try:
                frame = inspect.currentframe()
                lastframe = inspect.getouterframes(frame)[1][0]
                excepModule = inspect.getmodule(lastframe)
                if excepModule != None:
                    modName = excepModule.__name__
                    self.data['ModuleName'] = modName
            finally:
                frame = None

        # //
        # // Find out where the exception came from
        # //
        try:
            stack = inspect.stack(1)[1]
            self.data['FileName'] = stack[1]
            self.data['LineNumber'] = stack[2]
            self.data['MethodName'] = stack[3]
        finally:
            stack = None

        # //
        # // ClassName if ClassInstance is passed
        # //
        try:
            if self.data['ClassInstance'] != None:
                self.data['ClassName'] = self.data['ClassInstance'].__class__.__name__
        except Exception:
            pass

        # Determine the traceback at time of __init__
        try:
            self.traceback = "\n".join(traceback.format_tb(sys.exc_info()[2]))
        except Exception:
            self.traceback = "WMException error: Couldn't get traceback\n"

    def __getitem__(self, key):
        """
        make exception look like a dictionary
        """
        return self.data[key]

    def __setitem__(self, key, value):
        """
        make exception look like a dictionary
        """
        self.data[key] = value

    def addInfo(self, **data):
        """
        _addInfo_

        Add key=value information pairs to an
        exception instance
        """
        for key, value in viewitems(data):
            self[key] = value
        return

    def xml(self):
        """create a xml string rep of this exception"""
        strg = "<Exception>\n"
        strg += "<Object>\n"
        strg += "%s\n" % self.name
        strg += "</Object>\n"
        strg += "<Message>\n"
        strg += self._message
        strg += "</Message>\n"
        strg += "<DataItems>\n"
        for key, value in viewitems(self.data):
            strg += "<DataItem>\n"
            strg += "<Key>\n"
            strg += str(key)
            strg += "</Key>\n"
            strg += "<Value>\n"
            strg += str(value)
            strg += "</Value>\n"
            strg += "</DataItem>\n"
        strg += "</DataItems>\n"
        strg += "</Exception>\n"
        logging.error(strg)
        return strg

    def __str__(self):
        """create a string rep of this exception"""
        # WARNING: Do not change this string - it is used to extract error from log
        strg = WMEXCEPTION_START_STR
        strg += "\nException Class: %s\n" % self.name
        strg += "Message: %s\n" % self._message
        for key, value in viewitems(self.data):
            strg += "\t%s : %s\n" % (key, value,)
        strg += "\nTraceback: \n"
        strg += self.traceback
        strg += '\n'
        strg += WMEXCEPTION_END_STR
        if type(strg) == bytes:
            # Fix for the unicode encoding issue, #8043
            strg = strg.decode('utf-8', 'ignore')
        return strg

    def message(self):
        return self._message



#!/usr/bin/env python

"""
_Run_

container representing a run, and its constituent lumi sections and event counts
"""

from __future__ import print_function

from WMCore.DataStructs.WMObject import WMObject


class Run(WMObject):
    """
    _Run_

    Run container, is a list of lumi sections with associate event counts
    """

    def __init__(self, runNumber=None, *newLumis):
        WMObject.__init__(self)
        self.run = runNumber
        self.eventsPerLumi = {}
        self.extendLumis(newLumis)

    def __str__(self):
        return "Run%s:%s" % (self.run, self.eventsPerLumi)

    def __lt__(self, rhs):
        """
        Compare on run # first, then by lumis as a list is compared
        """
        if self.run != rhs.run:
            return self.run < rhs.run
        if sorted(self.eventsPerLumi.keys()) != sorted(rhs.eventsPerLumi.keys()):
            return sorted(self.eventsPerLumi.keys()) < sorted(rhs.eventsPerLumi.keys())
        return self.eventsPerLumi < rhs.eventsPerLumi

    def __gt__(self, rhs):
        """
        Compare on run # first, then by lumis as a list is compared
        """
        if self.run != rhs.run:
            return self.run > rhs.run
        if sorted(self.eventsPerLumi.keys()) != sorted(rhs.eventsPerLumi.keys()):
            return sorted(self.eventsPerLumi.keys()) > sorted(rhs.eventsPerLumi.keys())
        return self.eventsPerLumi > rhs.eventsPerLumi

    def extend(self, items):
        """
        Redirect to the function that already does this
        """
        self.extendLumis(items)
        return

    def __cmp__(self, rhs):
        return (self > rhs) - (self < rhs)  # Python3 equivalent of cmp()

    def __add__(self, rhs):
        """
        Combine two runs
        """
        if self.run != rhs.run:
            msg = "Adding together two different runs"
            msg += "Run %s does not equal Run %s" % (self.run, rhs.run)
            raise RuntimeError(msg)

        for lumi, events in rhs.eventsPerLumi.iteritems():
            if lumi not in self.eventsPerLumi or not self.eventsPerLumi[lumi]:  # Either doesn't exist, 0, or None
                self.eventsPerLumi[lumi] = events
            else:
                self.eventsPerLumi[lumi] += events
        return self

    def __iter__(self):
        return self.eventsPerLumi.__iter__()

    def __next__(self):
        """
        __next__ no longer needed
        """
        raise NotImplementedError

    def __len__(self):
        """
        Number of lumis
        """
        return self.eventsPerLumi.__len__()

    def __getitem__(self, key):
        """
        Get the nth lumi from the list (no event count)
        """
        return sorted(self.eventsPerLumi.keys()).__getitem__(key)

    def __setitem__(self, key, lumi):
        """
        Replace the nth lumi from the list (no event count)
        """
        try:
            oldLumi = sorted(self.eventsPerLumi.keys())[key]  # Extract the lumi from the sorted list
            del self.eventsPerLumi[oldLumi]  # Delete it and add the new one
        except IndexError:
            pass
        self.appendLumi(lumi)

    def __delitem__(self, key):
        try:
            oldLumi = sorted(self.eventsPerLumi.keys())[key]  # Extract the lumi from the sorted list
            del self.eventsPerLumi[oldLumi]  # Delete it
        except IndexError:
            pass

    def __eq__(self, rhs):
        """
        Check equality of run numbers and then underlying lumi/event dicts
        """
        if not isinstance(rhs, Run):
            return False
        if self.run != rhs.run:
            return False
        return self.eventsPerLumi == rhs.eventsPerLumi

    def __ne__(self, rhs):
        return not self.__eq__(rhs)

    def __hash__(self):
        """
        Calculate the value of the hash
        """
        value = self.run.__hash__()
        value += hash(frozenset(self.eventsPerLumi.items()))  # Hash that represents the dictionary
        return value

    @property
    def lumis(self):
        """
        Property that makes existing uses of myRun.lumis function by returning a list
        """
        return sorted(self.eventsPerLumi.keys())

    @lumis.setter
    def lumis(self, lumiList):
        """
        Setter to allow for replacement of the lumis with a list or list of tuples
        """
        self.eventsPerLumi = {}  # Remove existing dictionary
        for lumi in lumiList:
            if isinstance(lumi, (list, tuple)):
                self.eventsPerLumi[lumi[0]] = lumi[1]
            else:
                self.eventsPerLumi[lumi] = None

    def extendLumis(self, lumiList):
        """
        Method to replace myRun.lumis.extend() which does not work with the property
        """
        for lumi in lumiList:
            if not isinstance(lumi, (list, tuple)):  # comma separated lumi numbers
                self.eventsPerLumi[lumi] = None
            else:
                if isinstance(lumi, list) and not isinstance(lumi[0], tuple):  # then it's a plain list
                    for l in lumi:
                        self.eventsPerLumi[l] = None
                else:
                    if isinstance(lumi, tuple):  # it's an unpacked list of tuples
                        lumi = [(lumi)]
                    # it's a list/tuple of tuples
                    for tp in lumi:
                        if tp[0] in self.eventsPerLumi and self.eventsPerLumi[tp[0]]:
                            self.eventsPerLumi[tp[0]] += tp[1]  # Already exists, add events
                        else:  # Doesn't exist or is 0 or None
                            self.eventsPerLumi[tp[0]] = tp[1]

    def appendLumi(self, lumi):
        """
        Method to replace myRun.lumis.append() which does not work with the property
        """
        if isinstance(lumi, (list, tuple)) and self.eventsPerLumi[lumi[0]]:  # Already exists, add events
            self.eventsPerLumi[lumi[0]] += lumi[1]
        elif isinstance(lumi, (list, tuple)):  # Doesn't exist or is 0 or None
            self.eventsPerLumi[lumi[0]] = lumi[1]
        else:  # Just given lumis, not events
            if lumi not in self.eventsPerLumi:  # Don't overwrite existing events
                self.eventsPerLumi[lumi] = None

    def getEventsByLumi(self, lumi):
        """
        getter to select event counts by given lumi
        """
        return self.eventsPerLumi.get(lumi)

    def json(self):
        """
        _json_

        Convert to JSON friendly format.  Include some information for the
        thunker so that we can convert back.
        """
        return {"Run": self.run, "Lumis": self.eventsPerLumi,
                "thunker_encoded_json": True, "type": "WMCore.DataStructs.Run.Run"}

    def __to_json__(self, thunker=None):
        """
        __to_json__

        This is the standard way we JSONize other objects.
        Included here so we have a uniform method.
        """
        return self.json()

    def __from_json__(self, jsondata, thunker):
        """
        __from_json__

        Convert JSON data back into a Run object with integer lumi numbers
        """
        self.run = jsondata["Run"]
        self.eventsPerLumi = {}
        for lumi, events in jsondata["Lumis"].iteritems():
            self.eventsPerLumi[int(lumi)] = events  # Make the keys integers again

        return self
